@@ -15,10 +15,10 @@ from django.template import (
     Node,
     TemplateSyntaxError,
     Variable,
-    VariableDoesNotExist,
     loader,
 )
 from django.urls import reverse
+from django.urls.exceptions import NoReverseMatch
 from django.utils.module_loading import import_string
 from django.utils.safestring import mark_safe
 from django.utils.translation import gettext as _
@@ -263,9 +263,6 @@ def paginate_queryset(queryset, context, ckey_prefix):
 class BaseInkCommentNode(BaseCommentNode):
     def get_queryset(self, context):
         ctype, object_pk = self.get_target_ctype_pk(context)
-        if not object_pk:
-            return self.comment_model.objects.none()
-
         site_id = utils.get_current_site_id(context.get("request", None))
 
         kwargs = {
@@ -330,9 +327,9 @@ class RenderInkCommentListNode(InkCommentListNode):
         try:
             ctype, object_pk = self.get_target_ctype_pk(context)
         except AttributeError:
-            # in get_target_ctype_pk the call to FilterExpression.resolve does
-            # not raise VariableDoesNotExist, however in a latter step an
-            # AttributeError is raised when the object_expr does not exist
+            # in get_target_ctype_pk the call to FilterExpression.resolve
+            # does not raise VariableDoesNotExist, however in a latter step
+            # an AttributeError is raised when the object_expr does not exist
             # in the context. Therefore, this except raises when used as:
             # {% render_inkcomment_list for var_not_in_context %}
             return ""
@@ -359,11 +356,13 @@ class RenderInkCommentListNode(InkCommentListNode):
         app_model = "%s.%s" % (ctype.app_label, ctype.model)
         MTL = settings.COMMENTS_INK_MAX_THREAD_LEVEL_BY_APP_MODEL
         mtl = MTL.get(app_model, settings.COMMENTS_INK_MAX_THREAD_LEVEL)
+        max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
         context_dict.update(
             {
                 "dcx_theme_dir": theme_dir,
                 "max_thread_level": mtl,
                 "reply_stack": [],  # List to control reply widget rendering.
+                "max_users_in_tooltip": max_users_in_tooltip,
             }
         )
 
@@ -418,8 +417,6 @@ class InkCommentCountNode(BaseInkCommentNode):
 
     def render(self, context):
         ctype, object_pk = self.get_target_ctype_pk(context)
-        if not object_pk:
-            context[self.as_varname] = 0
         site_id = utils.get_current_site_id(context.get("request", None))
 
         result = None
@@ -427,7 +424,6 @@ class InkCommentCountNode(BaseInkCommentNode):
         key = settings.COMMENTS_INK_CACHE_KEYS["comments_count"].format(
             ctype_pk=ctype.pk, object_pk=object_pk, site_id=site_id
         )
-
         if dci_cache != None and key != "":
             cached = dci_cache.get(key)
             if cached:
@@ -449,19 +445,19 @@ class InkCommentCountNode(BaseInkCommentNode):
 def get_inkcomment_count(parser, token):
     """
     Gets the comment count for the given params and populates the template
-    context with a variable containing that value, whose name is defined by the
-    'as' clause.
+    context with a variable containing that value, whose name is defined by the 'as' clause. It differs from {% get_comment_count %} in which it
+    caches the result.
 
     Syntax::
 
-        {% get_comment_count for [object] as [varname]  %}
-        {% get_comment_count for [app].[model] [object_id] as [varname]  %}
+        {% get_inkcomment_count for [object] as [varname]  %}
+        {% get_inkcomment_count for [app].[model] [object_id] as [varname]  %}
 
     Example usage::
 
-        {% get_comment_count for event as comment_count %}
-        {% get_comment_count for calendar.event event.id as comment_count %}
-        {% get_comment_count for calendar.event 17 as comment_count %}
+        {% get_inkcomment_count for event as comment_count %}
+        {% get_inkcomment_count for calendar.event event.id as comment_count %}
+        {% get_inkcomment_count for calendar.event 17 as comment_count %}
 
     """
     return InkCommentCountNode.handle_token(parser, token)
@@ -718,7 +714,7 @@ def get_inkcomment_permalink(
         try:
             comments_folded = {int(cid) for cid in comments_folded.split(",")}
         except (TypeError, ValueError):
-            raise PageNotAnInteger(
+            raise TemplateSyntaxError(
                 _("A comment ID in comments_folded is not an integer.")
             )
 
@@ -981,20 +977,6 @@ def get_dci_theme_dir():
     return theme_dir
 
 
-@register.simple_tag(takes_context=True)
-def get_folded_comments_without(context, comment):
-    folded_comments = context.get("folded_comments", set())
-    folded_comments.remove(comment.id)
-    return ",".join([str(cm) for cm in folded_comments])
-
-
-@register.simple_tag(takes_context=True)
-def get_folded_comments_with(context, comment):
-    folded_comments = context.get("folded_comments", set())
-    folded_comments.add(comment.id)
-    return ",".join([str(cm) for cm in folded_comments])
-
-
 # ----------------------------------------------------------------------
 @register.inclusion_tag("comments/only_users_can_post.html")
 def render_only_users_can_post_template(object):
@@ -1047,10 +1029,10 @@ class BaseObjectReactionsNode(Node):
         if len(tokens) == 3:
             return cls(object_expr=parser.compile_filter(tokens[2]))
 
-        # {% render_object_reactions for app.models pk %}
+        # {% render_object_reactions for app.model pk %}
         elif len(tokens) == 4:
             return cls(
-                ctype=BaseCommentNode.lookup_content_type(tokens[2], tokens[0]),
+                ctype=cls.lookup_content_type(tokens[2], tokens[0]),
                 object_pk_expr=parser.compile_filter(tokens[3]),
             )
 
@@ -1071,11 +1053,6 @@ class BaseObjectReactionsNode(Node):
             )
 
     def __init__(self, ctype=None, object_pk_expr=None, object_expr=None):
-        if ctype is None and object_expr is None:
-            raise TemplateSyntaxError(
-                "Template tag must be given either a literal object "
-                "or a ctype and object pk."
-            )
         self.ctype = ctype
         self.object_pk_expr = object_pk_expr
         self.object_expr = object_expr
@@ -1091,27 +1068,33 @@ class BaseObjectReactionsNode(Node):
         self.site_id = utils.get_current_site_id(request)
 
         if self.object_expr:
-            try:
-                obj = self.object_expr.resolve(context)
-            except VariableDoesNotExist:
-                return
-            else:
+            obj = self.object_expr.resolve(context)
+            if obj:
                 self.ctype = ContentType.objects.get_for_model(obj)
                 self.object_pk = obj.pk
+            else:
+                raise TemplateSyntaxError(
+                    "Variable '%s' does not exist" % self.object_expr
+                )
         else:
             self.object_pk = self.object_pk_expr.resolve(
-                context, gnore_failures=True
+                context, ignore_failures=True
             )
 
         app_model = "%s.%s" % (self.ctype.app_label, self.ctype.model)
         options = utils.get_app_model_options(content_type=app_model)
-        if not "object_reactions_enabled" in options:
+        if not options["object_reactions_enabled"]:
             raise Exception(
                 "Object reactions for '%s' is not enabled in the setting "
                 "COMMENTS_INK_APP_MODEL_OPTIONS in the settings module"
                 % app_model
             )
 
+        self.object = self.ctype.get_object_for_this_type(pk=self.object_pk)
+        check_input_allowed_str = options.pop("check_input_allowed")
+        check_func = import_string(check_input_allowed_str)
+        self.is_input_allowed = check_func(self.object)
+
     @property
     def template_list(self):
         return [
@@ -1125,9 +1108,21 @@ class BaseObjectReactionsNode(Node):
 
     def get_object_reactions(self):
         """Returns list of dicts with object reactions and their counters."""
+        max_users_listed = getattr(
+            settings, "COMMENTS_INK_MAX_USERS_IN_TOOLTIP", 10
+        )
         reactionsd = dict(
             [
-                (item.reaction, item.counter)
+                (
+                    item.reaction,
+                    {
+                        "counter": item.counter,
+                        "authors": [
+                            settings.COMMENTS_INK_API_USER_REPR(author)
+                            for author in item.authors.all()[:max_users_listed]
+                        ],
+                    },
+                )
                 for item in ObjectReaction.objects.filter(
                     content_type=self.ctype,
                     object_pk=self.object_pk,
@@ -1137,17 +1132,15 @@ class BaseObjectReactionsNode(Node):
         )
 
         object_reactions = []
+        defs = {"counter": 0, "authors": []}
         for item in get_object_reactions_enum():
-            if item.value in reactionsd:
-                counter = reactionsd[item.value]
-            else:
-                counter = 0
             object_reactions.append(
                 {
                     "value": item.value,
                     "label": item.label,
                     "icon": item.icon,
-                    "counter": counter,
+                    "counter": reactionsd.get(item.value, defs)["counter"],
+                    "authors": reactionsd.get(item.value, defs)["authors"],
                 }
             )
 
@@ -1155,11 +1148,7 @@ class BaseObjectReactionsNode(Node):
 
     def render(self, context):
         request = context.get("request", None)
-        try:
-            self.resolve_ctype_and_object_pk(context)
-            object = self.ctype.get_object_for_this_type(pk=self.object_pk)
-        except AttributeError:
-            return ""
+        self.resolve_ctype_and_object_pk(context)
 
         cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
         page = (request and request.GET.get(cpage_qs_param, None)) or 1
@@ -1168,13 +1157,24 @@ class BaseObjectReactionsNode(Node):
         cfold = (request and request.GET.get(cfold_qs_param, None)) or None
         fold = (cfold and {int(cid) for cid in cfold.split(",")}) or {}
 
+        try:
+            login_url = reverse("login")
+        except NoReverseMatch as exc:
+            login_url = settings.LOGIN_URL
+
+        max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
+
         context = {
-            "object": object,
+            "content_type": self.ctype,
+            "object": self.object,
             "object_reactions": self.get_object_reactions(),
+            "is_input_allowed": self.is_input_allowed,
+            "max_users_in_tooltip": max_users_in_tooltip,
             "comments_page_qs_param": cpage_qs_param,
             cpage_qs_param: page,
             "comments_fold_qs_param": cfold_qs_param,
             cfold_qs_param: ",".join([str(cid) for cid in fold]),
+            "login_url": login_url,
         }
         htmlstr = loader.render_to_string(self.template_list, context, request)
         return htmlstr
@@ -1256,50 +1256,7 @@ def object_reactions_form_target(object):
         <form action="{% object_reactions_form_target object %}" method="post">
     """
     ctype = ContentType.objects.get_for_model(object)
-    return reverse("comments-ink-object-react", args=(ctype.id, object.id))
-
-
-class RenderObjectReactionsButtons(Node):
-    def __init__(self, user_reactions):
-        self.user_reactions = Variable(user_reactions)
-
-    def render(self, context):
-        context = {
-            "reactions": get_object_reactions_enum(),
-            "user_reactions": self.user_reactions.resolve(context),
-        }
-        template_list = [
-            pth.format(theme_dir=theme_dir)
-            for pth in _object_reactions_buttons_tmpl
-        ]
-        htmlstr = loader.render_to_string(template_list, context)
-        return htmlstr
-
-
-@register.tag
-def render_object_reactions_buttons(parser, token):
-    """
-    Renders template with reactions buttons, depending on the selected theme.
-
-    Example usage::
-
-        {% render_object_reactions_buttons user_reactions %}
-
-    Argument `user_reactions` is a list with `ReactionEnum` items, it
-    contains a user's reactions to an object. The template displays a button
-    per each reaction returned from get_object_reactions_enum().
-    Each reaction that is already in the `user_reactions` list is marked as
-    already clicked. This templatetag is used within the `react_to_object.html`
-    template.
-    """
-    try:
-        _, args = token.contents.split(None, 1)
-    except ValueError:
-        raise TemplateSyntaxError(
-            "%r tag requires argument" % token.contents.split()[0]
-        )
-    user_reactions = args
-    return RenderObjectReactionsButtons(user_reactions)
+    return reverse("comments-ink-react-to-object", args=(ctype.id, object.id))
 
 
 # ----------------------------------------------------------------------
