@@ -25,18 +25,19 @@ from django.utils.translation import gettext_lazy as _
 from django.views.decorators.csrf import csrf_protect
 from django.views.decorators.http import require_POST
 from django.views.defaults import bad_request
+
 from django_comments.signals import comment_was_posted, comment_will_be_posted
 from django_comments.views.comments import CommentPostBadRequest
 from django_comments.views.moderation import perform_flag
 from django_comments.views.utils import next_redirect
-from django_comments_ink import get_comment_reactions_enum, get_form
-from django_comments_ink import get_model as get_comment_model
+
 from django_comments_ink import (
+    get_comment_reactions_enum,
+    get_form,
     get_object_reactions_enum,
-    signals,
-    signed,
-    utils,
 )
+from django_comments_ink import get_model as get_comment_model
+from django_comments_ink import signals, signed, utils
 from django_comments_ink.conf import settings
 from django_comments_ink.models import (
     CommentReaction,
@@ -45,6 +46,7 @@ from django_comments_ink.models import (
     TmpInkComment,
 )
 
+
 InkComment = get_comment_model()
 
 
@@ -274,6 +276,45 @@ else:
     _reacted_tmpl = "comments/reacted.html"
 
 
+# List of possible paths to the list_reacted.html template file.
+_list_reacted_tmpl = []
+if theme_dir_exists:
+    _list_reacted_tmpl.extend(
+        [
+            "comments/{theme_dir}/{app_label}/{model}/list_reacted.html",
+            "comments/{theme_dir}/{app_label}/list_reacted.html",
+            "comments/{theme_dir}/list_reacted.html",
+        ]
+    )
+_list_reacted_tmpl.extend(
+    [
+        "comments/{app_label}/{model}/list_reacted.html",
+        "comments/{app_label}/list_reacted.html",
+        "comments/list_reacted.html",
+    ]
+)
+
+
+# List of possible paths to the list_reacted_to_object.html template file.
+_list_reacted_to_object_tmpl = []
+if theme_dir_exists:
+    _list_reacted_to_object_tmpl.extend(
+        [
+            "comments/{theme_dir}/{app_label}/{model}/"
+            "list_reacted_to_object.html",
+            "comments/{theme_dir}/{app_label}/list_reacted_to_object.html",
+            "comments/{theme_dir}/list_reacted_to_object.html",
+        ]
+    )
+_list_reacted_to_object_tmpl.extend(
+    [
+        "comments/{app_label}/{model}/list_reacted_to_object.html",
+        "comments/{app_label}/list_reacted_to_object.html",
+        "comments/list_reacted_to_object.html",
+    ]
+)
+
+
 # ---------------------------------------------------------------------
 @csrf_protect
 @require_POST
@@ -985,7 +1026,7 @@ def flag(request, comment_id, next=None):
         pk=comment_id,
         site__pk=utils.get_current_site_id(request),
     )
-    utils.check_option(comment, "comment_flagging_enabled")
+    utils.check_option("comment_flagging_enabled", comment=comment)
 
     # Flag on POST.
     if request.method == "POST":
@@ -1022,7 +1063,7 @@ def react(request, comment_id, next=None):
         pk=comment_id,
         site__pk=utils.get_current_site_id(request),
     )
-    utils.check_option(comment, "comment_reactions_enabled")
+    utils.check_option("comment_reactions_enabled", comment=comment)
     cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
     cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
 
@@ -1043,7 +1084,11 @@ def react(request, comment_id, next=None):
                 )
                 for pth in _reacted_js_tmpl
             ]
-            context = {"comment": comment}
+            max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
+            context = {
+                "comment": comment,
+                "max_users_in_tooltip": max_users_in_tooltip,
+            }
             status = 201 if created else 200
             return json_res(request, template_list, context, status=status)
 
@@ -1138,7 +1183,7 @@ def react_done(request):
             site__pk=utils.get_current_site_id(request),
         )
     else:
-        raise Http404
+        raise Http404(_("Comment does not exist"))
 
     page = int(cpage) if cpage != "last" else "last"
     fold = (len(cfold) and {int(cid) for cid in cfold.split(",")}) or {}
@@ -1168,10 +1213,12 @@ def react_to_object(request, content_type_id, object_pk, next=None):
             "does not exist" % (content_type_id, object_pk)
         )
 
+    ctype_str = ctype.app_label + "." + ctype.model
+    utils.check_option("object_reactions_enabled", content_type=ctype_str)
     cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
     cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
 
-    cpage = request.POST.get(cpage_qs_param, 1)
+    cpage = request.POST.get(cpage_qs_param, "")
     cfold = request.POST.get(cfold_qs_param, "")
 
     perform_react_to_object(request, content_type_id, object_pk)
@@ -1187,9 +1234,21 @@ def react_to_object(request, content_type_id, object_pk, next=None):
         )
 
     next = request.POST.get("next")
-    next, anchor = next.rsplit("#", 1)
-    joiner = ("?" in next) and "&" or "?"
-    next += joiner + "&".join(qs_params) + "#" + anchor
+
+    if "#" in next:
+        bits = next.rsplit("#", 1)
+        next = bits[0]
+        anchor = "#" + bits[1]
+    else:
+        anchor = ""
+
+    if len(qs_params):
+        joiner = ("?" in next) and "&" or "?"
+        next += joiner + "&".join(qs_params)
+
+    if len(anchor):
+        next += anchor
+
     return HttpResponseRedirect(next)
 
 
@@ -1206,6 +1265,7 @@ def perform_react_to_object(request, content_type_id, object_pk):
         object_pk=object_pk,
         site=site,
     )
+
     if or_qs.filter(authors=request.user).count() == 1:
         if or_qs[0].counter == 1:
             or_qs[0].delete_from_cache()
@@ -1234,6 +1294,74 @@ def perform_react_to_object(request, content_type_id, object_pk):
     return created
 
 
+def list_reacted(request, comment_id, reaction_value):
+    comment = get_object_or_404(
+        get_comment_model(),
+        pk=comment_id,
+        site__pk=utils.get_current_site_id(request),
+    )
+    reaction = get_object_or_404(
+        CommentReaction, reaction=reaction_value, comment=comment
+    )
+
+    authors = [
+        settings.COMMENTS_INK_API_USER_REPR(author)
+        for author in reaction.authors.all()
+    ]
+
+    max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
+    if len(authors) <= max_users_in_tooltip:
+        raise Http404(_("Not enough users"))
+
+    return render(
+        request,
+        _list_reacted_tmpl,
+        {
+            "comment": comment,
+            "reaction": get_comment_reactions_enum()(reaction.reaction),
+            "authors": authors,
+        },
+    )
+
+
+def list_reacted_to_object(request, content_type_id, object_pk, reaction_value):
+    try:
+        ctype = ContentType.objects.get(pk=content_type_id)
+        ctype.get_object_for_this_type(pk=object_pk)
+    except Exception:
+        raise Http404(
+            "Object referenced by pair (ctype_id, obj_id): (%d, %d) "
+            "does not exist" % (content_type_id, object_pk)
+        )
+
+    reaction = get_object_or_404(
+        ObjectReaction,
+        reaction=reaction_value,
+        content_type=ctype,
+        object_pk=object_pk,
+        site=get_current_site(request),
+    )
+
+    authors = [
+        settings.COMMENTS_INK_API_USER_REPR(author)
+        for author in reaction.authors.all()
+    ]
+
+    max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
+    if len(authors) <= max_users_in_tooltip:
+        raise Http404(_("Not enough users"))
+
+    return render(
+        request,
+        _list_reacted_to_object_tmpl,
+        {
+            "object": reaction.content_object,
+            "reaction": get_object_reactions_enum()(reaction.reaction),
+            "authors": authors,
+        },
+    )
+
+
 def get_inkcomment_url(request, content_type_id, object_id, comment_id):
     response = shortcut(request, content_type_id, object_id)
 
